searchNodes=[{"doc":"Client-driven interactive widgets for Livebook. Kino is the library used by Livebook to render rich and interactive outputs directly from your Elixir code. Built-in widgets Kino renders any data structure that implements the Kino.Render protocol, falling back to the Kernel.inspect/2 representation whenever an implementation is not available. The data structures supported by Kino out of the box are: VegaLite VegaLite specifications are rendered as visualizations: Vl . new ( ... ) |&gt; Vl . data_from_series ( ... ) |&gt; ... Kino.VegaLite Kino.VegaLite is an extension of VegaLite that allows data to be streamed: widget = Vl . new ( ... ) |&gt; Vl . data_from_series ( ... ) |&gt; ... |&gt; Kino.VegaLite . new ( ) |&gt; Kino . render ( ) Kino.VegaLite . push ( widget , %{ x : 1 , y : 2 } ) Kino.ETS Kino.ETS implements a data table output for ETS tables in the system: tid = :ets . new ( :users , [ :set , :public ] ) Kino.ETS . new ( tid ) Kino.DataTable Kino.DataTable implements a data table output for user-provided tabular data: data = [ %{ id : 1 , name : &quot;Elixir&quot; , website : &quot;https://elixir-lang.org&quot; } , %{ id : 2 , name : &quot;Erlang&quot; , website : &quot;https://www.erlang.org&quot; } ] Kino.DataTable . new ( data ) Kino.Image Kino.Image wraps binary image content and can be used to render raw images of any given format: content = File . read! ( &quot;/path/to/image.jpeg&quot; ) Kino.Image . new ( content , &quot;image/jpeg&quot; ) Kino.Markdown Kino.Markdown wraps Markdown content for richer text rendering. Kino.Markdown . new ( &quot;&quot;&quot; # Example A regular Markdown file. ## Code ```elixir &quot;Elixir&quot; |&gt; String.graphemes() |&gt; Enum.frequencies() ``` ## Table | ID | Name | Website | | -- | ------ | ----------------------- | | 1 | Elixir | https://elixir-lang.org | | 2 | Erlang | https://www.erlang.org | &quot;&quot;&quot; ) Kino.Ecto Kino.Ecto implements a data table output for arbitrary Ecto queries: Kino.Ecto . new ( Weather , Repo ) Kino.Frame Kino.Frame is a placeholder for static outptus that can be dynamically updated. widget = Kino.Frame . new ( ) |&gt; Kino . render ( ) for i &lt;- 1 .. 100 do Kino.Frame . render ( widget , i ) Process . sleep ( 50 ) end Also see Kino.animate/3 . User interactions Kino.Input and Kino.Control provide a set of widgets for entering data and capturing user events. See the respective module documentation for examples. All others All other data structures are rendered as text using Elixir's Kernel.inspect/2 . Custom widgets Kino makes it possible to define custom JavaScript powered widgets, see Kino.JS and Kino.JS.Live for more details.","ref":"Kino.html","title":"Kino","type":"module"},{"doc":"Returns a widget that periodically calls the given function to render a new result. The callback is run every interval_ms milliseconds and receives the accumulated value. The callback should return either of: {:cont, term_to_render, acc} - the continue :halt - to no longer schedule callback evaluation This function uses Kino.Frame as the underlying widget. Examples # Render new Markdown every 100ms Kino . animate ( 100 , 0 , fn i -&gt; md = Kino.Markdown . new ( &quot;**Iteration: ` \#{ i } `**&quot; ) { :cont , md , i + 1 } end )","ref":"Kino.html#animate/3","title":"Kino.animate/3","type":"function"},{"doc":"Configures Kino. The supported options are: :inspect They are discussed individually in the sections below. Inspect A keyword list containing inspect options used for printing usual evaluation results. Defaults to pretty formatting with a limit of 50 entries. To show more entries, you configure a higher limit: Kino . configure ( inspect : [ limit : 200 ] ) You can also show all entries by setting the limit to :infinity , but keep in mind that for large data structures it is memory-expensive and is not an advised configuration in this case. Instead prefer the use of IO.inspect/2 with :infinity limit when needed. See Inspect.Opts for the full list of options.","ref":"Kino.html#configure/1","title":"Kino.configure/1","type":"function"},{"doc":"Inspects the given term as cell output. This works essentially the same as IO.inspect/2 , except it always produces colored text and respects the configuration set with configure/1 . Opposite to render/1 , it does not attempt to render the given term as a widget.","ref":"Kino.html#inspect/2","title":"Kino.inspect/2","type":"function"},{"doc":"Returns a special value that results in no visible output.","ref":"Kino.html#nothing/0","title":"Kino.nothing/0","type":"function"},{"doc":"Renders the given term as cell output. This effectively allows any Livebook cell to have multiple evaluation results.","ref":"Kino.html#render/1","title":"Kino.render/1","type":"function"},{"doc":"Starts a process under the Kino supervisor. The process is automatically terminated when the current process terminates or the current cell reevaluates.","ref":"Kino.html#start_child/1","title":"Kino.start_child/1","type":"function"},{"doc":"","ref":"Kino.html#t:nothing/0","title":"Kino.nothing/0","type":"type"},{"doc":"A widget for interactively viewing enumerable data. The data must be an enumerable of records, where each record is either map, struct, keyword list or tuple. Examples data = [ %{ id : 1 , name : &quot;Elixir&quot; , website : &quot;https://elixir-lang.org&quot; } , %{ id : 2 , name : &quot;Erlang&quot; , website : &quot;https://www.erlang.org&quot; } ] Kino.DataTable . new ( data ) The tabular view allows you to quickly preview the data and analyze it thanks to sorting capabilities. data = Process . list ( ) |&gt; Enum . map ( &amp; Process . info / 1 ) Kino.DataTable . new ( data , keys : [ :registered_name , :initial_call , :reductions , :stack_size ] )","ref":"Kino.DataTable.html","title":"Kino.DataTable","type":"module"},{"doc":"Starts a widget process with enumerable tabular data. Options :keys - a list of keys to include in the table for each record. The order is reflected in the rendered table. For tuples use 0-based indices. Optional. :sorting_enabled - whether the widget should support sorting the data. Sorting requires traversal of the whole enumerable, so it may not be desirable for lazy enumerables. Defaults to true if data is a list and false otherwise. :show_underscored - whether to include record keys starting with underscore. This option is ignored if :keys is also given. Defaults to false .","ref":"Kino.DataTable.html#new/2","title":"Kino.DataTable.new/2","type":"function"},{"doc":"","ref":"Kino.DataTable.html#t:t/0","title":"Kino.DataTable.t/0","type":"type"},{"doc":"A widget for interactively viewing an ETS table. Examples tid = :ets . new ( :users , [ :set , :public ] ) Kino.ETS . new ( tid ) Kino.ETS . new ( :elixir_config )","ref":"Kino.ETS.html","title":"Kino.ETS","type":"module"},{"doc":"Starts a widget process representing the given ETS table. Note that private tables cannot be read by an arbitrary process, so the given table must have either public or protected access.","ref":"Kino.ETS.html#new/1","title":"Kino.ETS.new/1","type":"function"},{"doc":"","ref":"Kino.ETS.html#t:t/0","title":"Kino.ETS.t/0","type":"type"},{"doc":"A widget for interactively viewing Ecto query results. The data must be an enumerable of records, where each record is either map, struct, keyword list or tuple. Examples The widget primarily allows for viewing a database table given a schema: Kino.Ecto . new ( Weather , Repo ) However, the first argument can be any queryable, so you can pipe arbitrary queries directly to the widget: from ( w in Weather , where : w . city == &quot;New York&quot; ) |&gt; Kino.Ecto . new ( Repo )","ref":"Kino.Ecto.html","title":"Kino.Ecto","type":"module"},{"doc":"Starts a widget process with the given queryable as the data source.","ref":"Kino.Ecto.html#new/2","title":"Kino.Ecto.new/2","type":"function"},{"doc":"","ref":"Kino.Ecto.html#t:t/0","title":"Kino.Ecto.t/0","type":"type"},{"doc":"A placeholder for outputs. A frame wraps outputs that can be dynamically updated at any time. Also see Kino.animate/3 which offers a convenience on top of this widget. Examples widget = Kino.Frame . new ( ) |&gt; Kino . render ( ) for i &lt;- 1 .. 100 do Kino.Frame . render ( widget , i ) Process . sleep ( 50 ) end Or with a scheduled task in the background. widget = Kino.Frame . new ( ) |&gt; Kino . render ( ) Kino.Frame . periodically ( widget , 50 , 0 , fn i -&gt; Kino.Frame . render ( widget , i ) { :cont , i + 1 } end )","ref":"Kino.Frame.html","title":"Kino.Frame","type":"module"},{"doc":"Renders and appends the given term to the frame.","ref":"Kino.Frame.html#append/2","title":"Kino.Frame.append/2","type":"function"},{"doc":"Removes all outputs within the given frame.","ref":"Kino.Frame.html#clear/1","title":"Kino.Frame.clear/1","type":"function"},{"doc":"Starts a widget process.","ref":"Kino.Frame.html#new/0","title":"Kino.Frame.new/0","type":"function"},{"doc":"Registers a callback to run periodically in the widget process. The callback is run every interval_ms milliseconds and receives the accumulated value. The callback should return either of: {:cont, acc} - the continue with the new accumulated value :halt - to no longer schedule callback evaluation The callback is run for the first time immediately upon registration.","ref":"Kino.Frame.html#periodically/4","title":"Kino.Frame.periodically/4","type":"function"},{"doc":"Renders the given term within the frame. This works similarly to Kino.render/1 , but the rendered output replaces existing frame contents.","ref":"Kino.Frame.html#render/2","title":"Kino.Frame.render/2","type":"function"},{"doc":"","ref":"Kino.Frame.html#t:t/0","title":"Kino.Frame.t/0","type":"opaque"},{"doc":"A struct wrapping a binary image. This is just a meta-struct that implements the Kino.Render protocol, so that it gets rendered as the underlying image. Examples content = File . read! ( &quot;/path/to/image.jpeg&quot; ) Kino.Image . new ( content , &quot;image/jpeg&quot; )","ref":"Kino.Image.html","title":"Kino.Image","type":"module"},{"doc":"Wraps the given binary content into the image struct. The given type be either :jpeg , :png , :gif , :svg or a string with image MIME type.","ref":"Kino.Image.html#new/2","title":"Kino.Image.new/2","type":"function"},{"doc":"","ref":"Kino.Image.html#t:common_image_type/0","title":"Kino.Image.common_image_type/0","type":"type"},{"doc":"","ref":"Kino.Image.html#t:mime_type/0","title":"Kino.Image.mime_type/0","type":"type"},{"doc":"","ref":"Kino.Image.html#t:t/0","title":"Kino.Image.t/0","type":"opaque"},{"doc":"A struct wrapping a Markdown content. This is just a meta-struct that implements the Kino.Render protocol, so that it gets rendered as markdown. Examples Kino.Markdown . new ( &quot;&quot;&quot; # Example A regular Markdown file. ## Code ```elixir &quot;Elixir&quot; |&gt; String.graphemes() |&gt; Enum.frequencies() ``` ## Table | ID | Name | Website | | -- | ------ | ----------------------- | | 1 | Elixir | https://elixir-lang.org | | 2 | Erlang | https://www.erlang.org | &quot;&quot;&quot; ) This format may come in handy when exploring Markdown from external sources: content = File . read! ( &quot;/path/to/README.md&quot; ) Kino.Markdown . new ( content )","ref":"Kino.Markdown.html","title":"Kino.Markdown","type":"module"},{"doc":"Wraps the given binary content into the markdown struct.","ref":"Kino.Markdown.html#new/1","title":"Kino.Markdown.new/1","type":"function"},{"doc":"","ref":"Kino.Markdown.html#t:t/0","title":"Kino.Markdown.t/0","type":"opaque"},{"doc":"A widget wrapping VegaLite graphic. This widget allow for rendering regular VegaLite graphic and then streaming new data points to update the graphic. Examples widget = Vl . new ( width : 400 , height : 400 ) |&gt; Vl . mark ( :line ) |&gt; Vl . encode_field ( :x , &quot;x&quot; , type : :quantitative ) |&gt; Vl . encode_field ( :y , &quot;y&quot; , type : :quantitative ) |&gt; Kino.VegaLite . new ( ) |&gt; Kino . render ( ) for i &lt;- 1 .. 300 do point = %{ x : i / 10 , y : :math . sin ( i / 10 ) } Kino.VegaLite . push ( widget , point ) Process . sleep ( 25 ) end","ref":"Kino.VegaLite.html","title":"Kino.VegaLite","type":"module"},{"doc":"Removes all data points from the graphic dataset. Options dataset - name of the targetted dataset from the VegaLite specification. Defaults to the default anonymous dataset.","ref":"Kino.VegaLite.html#clear/2","title":"Kino.VegaLite.clear/2","type":"function"},{"doc":"Starts a widget process with the given VegaLite definition.","ref":"Kino.VegaLite.html#new/1","title":"Kino.VegaLite.new/1","type":"function"},{"doc":"Registers a callback to run periodically in the widget process. The callback is run every interval_ms milliseconds and receives the accumulated value. The callback should return either of: {:cont, acc} - the continue with the new accumulated value :halt - to no longer schedule callback evaluation The callback is run for the first time immediately upon registration.","ref":"Kino.VegaLite.html#periodically/4","title":"Kino.VegaLite.periodically/4","type":"function"},{"doc":"Appends a single data point to the graphic dataset. Options :window - the maximum number of data points to keep. This option is useful when you are appending new data points to the plot over a long period of time. dataset - name of the targetted dataset from the VegaLite specification. Defaults to the default anonymous dataset.","ref":"Kino.VegaLite.html#push/3","title":"Kino.VegaLite.push/3","type":"function"},{"doc":"Appends a number of data points to the graphic dataset. See push/3 for more details.","ref":"Kino.VegaLite.html#push_many/3","title":"Kino.VegaLite.push_many/3","type":"function"},{"doc":"","ref":"Kino.VegaLite.html#t:t/0","title":"Kino.VegaLite.t/0","type":"type"},{"doc":"Various widgets for user interactions. Each widget is a UI control element that the user interacts with, consequenty producing an event stream. Those widgets are often useful paired with Kino.Frame for presenting content that changes upon user interactions. Examples First, create a control and make sure it is rendered, either by placing it at the end of a code cell or by explicitly rendering it with Kino.render/1 . button = Kino.Control . button ( &quot;Hello&quot; ) Next, to receive events from the control, a process needs to subscribe to it and specify pick a name to distinguish the events. Kino.Control . subscribe ( button , :hello ) As the user interacts with the button, the subscribed process receives corresponding events. IEx.Helpers . flush ( ) #=&gt; {:hello, %{origin: #PID&lt;10895.9854.0&gt;}} #=&gt; {:hello, %{origin: #PID&lt;10895.9854.0&gt;}}","ref":"Kino.Control.html","title":"Kino.Control","type":"module"},{"doc":"Creates a new button.","ref":"Kino.Control.html#button/1","title":"Kino.Control.button/1","type":"function"},{"doc":"Creates a new form. A form is composed of regular inputs from the Kino.Input module, however in a form input values are not synchronized between users. Consequently, the form is another control for producing user-specific events. Either :submit or :report_changes must be specified. Options :submit - specifies the label to use for the submit button and enables submit events :report_changes - whether to send new form value whenever any of the input changes. Defaults to false :reset_on_submit - a list of fields to revert to their default values once the form is submitted. Use true to indicate all fields. Defaults to [] Event info In addition to standard properties, all events include additional properties. :type - either :submit or :change :data - a map with field values, matching the field list Examples Create a form out of inputs: form = Kino.Control . form ( [ name : Kino.Input . text ( &quot;Name&quot; ) , message : Kino.Input . textarea ( &quot;Message&quot; ) ] , submit : &quot;Send&quot; ) Subscribe to events: Kino.Control . subscribe ( form , :chat_form ) As users submit the form the payload is sent: IEx.Helpers . flush ( ) #=&gt; {:chat_form, #=&gt; %{ #=&gt; data: %{message: &quot;Hola&quot;, name: &quot;Amy&quot;}, #=&gt; origin: #PID&lt;10905.5195.0&gt;, #=&gt; type: :submit #=&gt; }} #=&gt; {:chat_form, #=&gt; %{ #=&gt; data: %{message: &quot;Hey!&quot;, name: &quot;Jake&quot;}, #=&gt; origin: #PID&lt;10905.5186.0&gt;, #=&gt; type: :submit #=&gt; }}","ref":"Kino.Control.html#form/2","title":"Kino.Control.form/2","type":"function"},{"doc":"Returns a new interval event source. This can be used as event source for stream/1 and tagged_stream/1 . The events are emitted periodically with an increasing value, starting from 0 and have the form: %{ type : :interval , iteration : non_neg_integer ( ) }","ref":"Kino.Control.html#interval/1","title":"Kino.Control.interval/1","type":"function"},{"doc":"Creates a new keyboard control. This widget is represented as button that toggles interception mode, in which the given keyboard events are captured. Event info In addition to standard properties, all events include additional properties. Key events :type - either :keyup or :keydown :key - the value matching the browser KeyboardEvent.key Status event :type - either :status :enabled - whether the keyboard is activated Examples Create the widget: keyboard = Kino.Control . keyboard ( [ :keyup , :keydown , :status ] ) Subscribe to events: Kino.Control . subscribe ( keyboard , :keyboard ) As the user types events are streamed: IEx.Helpers . flush ( ) #=&gt; {:keyboard, %{enabled: true, origin: #PID&lt;10895.9854.0&gt;, type: :status} #=&gt; {:keyboard, %{key: &quot;o&quot;, origin: #PID&lt;10895.9854.0&gt;, type: :keydown}} #=&gt; {:keyboard, %{key: &quot;k&quot;, origin: #PID&lt;10895.9854.0&gt;, type: :keydown}} #=&gt; {:keyboard, %{key: &quot;o&quot;, origin: #PID&lt;10895.9854.0&gt;, type: :keyup}} #=&gt; {:keyboard, %{key: &quot;k&quot;, origin: #PID&lt;10895.9854.0&gt;, type: :keyup}}","ref":"Kino.Control.html#keyboard/1","title":"Kino.Control.keyboard/1","type":"function"},{"doc":"Returns a Stream of control events. This is an alternative API to subscribe/2 , such that event messages are consume via stream instead of process messages. It accepts a single source or a list of sources, where each source is either of: Kino.Control - emitting value on relevant interaction Kino.Input - emitting value on value change Example button = Kino.Control . button ( &quot;Hello&quot; ) for event &lt;- Kino.Control . stream ( button ) do IO . inspect ( event ) end #=&gt; %{origin: #PID&lt;10895.9854.0&gt;, type: :click} #=&gt; %{origin: #PID&lt;10895.9854.0&gt;, type: :click} Or with multiple sources: button = Kino.Control . button ( &quot;Hello&quot; ) input = Kino.Input . checkbox ( &quot;Check&quot; ) interval = Kino.Control . interval ( 1000 ) for event &lt;- Kino.Control . stream ( [ button , input , interval ] ) do IO . inspect ( event ) end #=&gt; %{type: :interval, iteration: 0} #=&gt; %{origin: #PID&lt;10895.9854.0&gt;, type: :click} #=&gt; %{origin: #PID&lt;10895.9854.0&gt;, type: :change, value: true}","ref":"Kino.Control.html#stream/1","title":"Kino.Control.stream/1","type":"function"},{"doc":"Subscribes the calling process to control or input events. The events are sent as {tag, info} , where info is a map with event details. In particular, it always includes :origin , which is an opaque identifier of the client that triggered the event.","ref":"Kino.Control.html#subscribe/2","title":"Kino.Control.subscribe/2","type":"function"},{"doc":"Same as stream/1 , but attaches custom tag to every stream item. Example button = Kino.Control . button ( &quot;Hello&quot; ) input = Kino.Input . checkbox ( &quot;Check&quot; ) for event &lt;- Kino.Control . tagged_stream ( [ hello : button , check : input ] ) do IO . inspect ( event ) end #=&gt; {:hello, %{origin: #PID&lt;10895.9854.0&gt;, type: :click}} #=&gt; {:check, %{origin: #PID&lt;10895.9854.0&gt;, type: :change, value: true}}","ref":"Kino.Control.html#tagged_stream/1","title":"Kino.Control.tagged_stream/1","type":"function"},{"doc":"Unsubscribes the calling process from control or input events.","ref":"Kino.Control.html#unsubscribe/1","title":"Kino.Control.unsubscribe/1","type":"function"},{"doc":"","ref":"Kino.Control.html#t:interval/0","title":"Kino.Control.interval/0","type":"opaque"},{"doc":"","ref":"Kino.Control.html#t:t/0","title":"Kino.Control.t/0","type":"opaque"},{"doc":"Various input elements for entering data. Examples First, create an input and make sure it is rendered, either by placing it at the end of a code cell or by explicitly rendering it with Kino.render/1 . input = Kino.Input . text ( &quot;Name&quot; ) Then read the value at any later point: name = Kino.Input . read ( input ) Async API You can subscribe to input changes or use the Stream API for event feed. See the Kino.Control module for more details.","ref":"Kino.Input.html","title":"Kino.Input","type":"module"},{"doc":"Creates a new checkbox. The input value can be either true or false . Options :default - the initial input value. Defaults to false","ref":"Kino.Input.html#checkbox/2","title":"Kino.Input.checkbox/2","type":"function"},{"doc":"Creates a new color input. The input value can be a hex color string. Options :default - the initial input value. Defaults to #6583FF","ref":"Kino.Input.html#color/2","title":"Kino.Input.color/2","type":"function"},{"doc":"Creates a new number input. The input value is can be either a number or nil . Options :default - the initial input value. Defaults to nil","ref":"Kino.Input.html#number/2","title":"Kino.Input.number/2","type":"function"},{"doc":"Creates a new password input. This is similar to text input, except the content is not visible by default. Options :default - the initial input value. Defaults to &quot;&quot;","ref":"Kino.Input.html#password/2","title":"Kino.Input.password/2","type":"function"},{"doc":"Creates a new slider input. The input value can be either float in the configured range. Options :default - the initial input value. Defaults to the minimum value :min - the minimum value :max - the maximum value :step - the slider increment","ref":"Kino.Input.html#range/2","title":"Kino.Input.range/2","type":"function"},{"doc":"Synchronously reads the current input value. Note that to retrieve the value, the input must be rendered first, otherwise an error is raised. Examples input = Kino.Input . text ( &quot;Name&quot; ) Kino.Input . read ( input )","ref":"Kino.Input.html#read/1","title":"Kino.Input.read/1","type":"function"},{"doc":"Creates a new select input. The input expects a list of options in the form [{value, label}] , where value is an arbitrary term and label is a descriptive string. Options :default - the initial input value. Defaults to the first value from the given list of options Examples Kino.Input . select ( &quot;Language&quot; , [ en : &quot;English&quot; , fr : &quot;Français&quot; ] ) Kino.Input . select ( &quot;Language&quot; , [ { 1 , &quot;One&quot; } , { 2 , &quot;Two&quot; } , { 3 , &quot;Three&quot; } ] )","ref":"Kino.Input.html#select/3","title":"Kino.Input.select/3","type":"function"},{"doc":"Creates a new text input. Options :default - the initial input value. Defaults to &quot;&quot;","ref":"Kino.Input.html#text/2","title":"Kino.Input.text/2","type":"function"},{"doc":"Creates a new multiline text input. Options :default - the initial input value. Defaults to &quot;&quot;","ref":"Kino.Input.html#textarea/2","title":"Kino.Input.textarea/2","type":"function"},{"doc":"Creates a new URL input. The input value can be either a valid URL string or nil . Options :default - the initial input value. Defaults to nil","ref":"Kino.Input.html#url/2","title":"Kino.Input.url/2","type":"function"},{"doc":"","ref":"Kino.Input.html#t:t/0","title":"Kino.Input.t/0","type":"opaque"},{"doc":"Allows for defining custom JavaScript powered widgets. Example Here's how we could define a minimal widget that embeds the given HTML directly into the page. defmodule Kino.HTML do use Kino.JS def new ( html ) do Kino.JS . new ( __MODULE__ , html ) end asset &quot;main.js&quot; do &quot;&quot;&quot; export function init(ctx, html) { ctx.root.innerHTML = html; } &quot;&quot;&quot; end end Let's break down the API. To define a custom widget we need to create a new module, conventionally under the Kino. prefix, so that the end user can easily autocomplete all available widgets. In this case we go with Kino.HTML . We start by adding use Kino.JS , which makes our module asset-aware. In particular, it allows us to use the asset/2 macro to define arbitrary files directly in the module source. All widgets require a main.js file that defines a JavaScript module and becomes the entrypoint on the client side. The JavaScript module is expected to export the init(ctx, data) function, where ctx is a special object (discussed in detail later) and data is the widget data passed from the Elixir side. In our example the init function accesses the root element with ctx.root and overrides its content with the given HTML string. Finally, we define the new(html) function that builds widgets with the given HTML. Underneath we call Kino.JS.new/2 specifying our module as the widget type and passing the data (available in the JavaScript init function later). Again, it's a convention for each widget module to define a new function to provide uniform experience for the end user. Assets We already saw how to define a JavaScript (or any other) file using the asset/2 macro, however in most cases it's preferable to put assets in a dedicated directory to benefit from syntax highlighting and other editor features. To do that, we just need to specify where the corresponding directory is located: use Kino.JS , assets_path : &quot;lib/assets/html&quot; URLs When using multiple asset files, make sure to use relative URLs. For example, when adding an image to the page, instead of: &lt; img src = &quot;/images/cat.jpeg&quot; / &gt; Do: &lt; img src = &quot;./images/cat.jpeg&quot; / &gt; This will correctly point to the images/cat.jpeg file in your assets. JavaScript API In the example we briefly introduced the ctx (context) object that is made available in the init(ctx, data) function. This object encapsulates all of the Livebook-specific API that we can call on the JavaScript side. Properties ctx.root - the root element controlled by the widget Functions ctx.importCSS(url) - loads CSS from the given URL into the page. Returns a Promise that resolves once the CSS is loaded ctx.handleEvent(event, callback) - registers an event handler. Once event is broadcasted, callback is executed with the event payload. This applies to Kino.JS.Live widgets ctx.pushEvent(event, payload) - sends an event to the widget server, where it is handled with Kino.JS.Live.handle_event/3 . This applies to Kino.JS.Live widgets CDN It is possible to use a regular JavaScript bundler for generating the assets, however in many cases a simpler and preferred approach is to import the necessary dependencies directly from a CDN. To give a concrete example, here's how we could use the mermaid JavaScript package for rendering diagrams: defmodule Kino.Mermaid do use Kino.JS def new ( graph ) do Kino.JS . new ( __MODULE__ , graph ) end asset &quot;main.js&quot; do &quot;&quot;&quot; import &quot;https://cdn.jsdelivr.net/npm/mermaid@8.13.3/dist/mermaid.min.js&quot;; mermaid.initialize({ startOnLoad: false }); export function init(ctx, graph) { mermaid.render(&quot;graph1&quot;, graph, (svgSource, bindListeners) =&gt; { ctx.root.innerHTML = svgSource; bindListeners &amp;&amp; bindListeners(ctx.root); }); } &quot;&quot;&quot; end end And we would use it like so: Kino.Mermaid . new ( &quot;&quot;&quot; graph TD; A--&gt;B; A--&gt;C; B--&gt;D; C--&gt;D; &quot;&quot;&quot; ) Live widgets So far we covered the API for defining static widgets, where the JavaScript side only receives the initial data and there is no further interaction with the Elixir side. To introduce such interaction, see Kino.JS.Live as a next step in our discussion.","ref":"Kino.JS.html","title":"Kino.JS","type":"module"},{"doc":"Defines an asset file. This serves as a convenience when prototyping or building simple widgets, otherwise you most likely want to put assets in separate files. See the Assets for more details. Examples asset &quot;main.js&quot; do &quot;&quot;&quot; export function init(ctx, data) { ... } &quot;&quot;&quot; end asset &quot;main.css&quot; do &quot;&quot;&quot; .box { ... } &quot;&quot;&quot; end","ref":"Kino.JS.html#asset/2","title":"Kino.JS.asset/2","type":"macro"},{"doc":"Instantiates a static JavaScript widget defined by module . The given data is passed directly to the JavaScript side during initialization. Options :export_info_string - used as the info string for the Markdown code block where output data is persisted :export_key - in case the data is a map and only a specific part should be exported Export The output can optionally be exported in notebook source by specifying :export_info_string . For example: data = &quot;graph TD;A--&gt;B;&quot; Kino.JS . new ( __MODULE__ , data , export_info_string : &quot;mermaid&quot; ) Would be rendered as the following Live Markdown: ```mermaid graph TD;A--&gt;B; ``` Non-binary data is automatically serialized to JSON.","ref":"Kino.JS.html#new/3","title":"Kino.JS.new/3","type":"function"},{"doc":"","ref":"Kino.JS.html#t:t/0","title":"Kino.JS.t/0","type":"opaque"},{"doc":"Introduces state and event-driven capabilities to JavaScript powered widgets. Make sure to read the introduction to JavaScript widgets in Kino.JS for more context. Similarly to static widgets, live widgets involve a custom JavaScript code running in the browser. In fact, this part of the API is the same. In addition, each live widget has a server process running on the Elixir side, responsible for maintaining state and able to communicate with the JavaScript side at any time. Again, to illustrate the ideas we start with a minimal example. Example We will follow up on our Kino.HTML example by adding support for replacing the content on demand. defmodule Kino.LiveHTML do use Kino.JS use Kino.JS.Live def new ( html ) do Kino.JS.Live . new ( __MODULE__ , html ) end def replace ( widget , html ) do Kino.JS.Live . cast ( widget , { :replace , html } ) end @impl true def init ( html , ctx ) do { :ok , assign ( ctx , html : html ) } end @impl true def handle_connect ( ctx ) do { :ok , ctx . assigns . html , ctx } end @impl true def handle_cast ( { :replace , html } , ctx ) do broadcast_event ( ctx , &quot;replace&quot; , html ) { :noreply , assign ( ctx , html : html ) } end asset &quot;main.js&quot; do &quot;&quot;&quot; export function init(ctx, html) { ctx.root.innerHTML = html; ctx.handleEvent(&quot;replace&quot;, (html) =&gt; { ctx.root.innerHTML = html; }); } &quot;&quot;&quot; end end Just as before we define a module, this time calling it Kino.LiveHTML for clarity. Note many similarities to the previous version, we still call use Kino.JS , define the main.js file and define the new(html) function for building the widget. As a matter of fact, the initial result of Kino.LiveHTML.new(html) will render exactly the same as our previous Kino.HTMl.new(html) . As for the new bits, we added use Kino.JS.Live to define a live widget server. We use Kino.JS.Live.new/2 for creating the widget instance and we implement a few GenServer -like callbacks. Once the widget server is started with Kino.JS.Live.new/2 , the init/2 callback is called with the initial argument. In this case we store the given html in server state. Whenever the widget is rendered on a new client, the handle_connect/1 callback is called and it builds the initial data for the client. In this case, we always return the stored html . This initial data is then passed to the JavaScript init function. Keep in mind that while the server is initialized once, connect may happen at any point, as the users join/refresh the page. Finally, the whole point of our example is the ability to replace the HTML content directly from the Elixir side and for this purpose we added the public replace(widget, html) function. Underneath the function uses cast/2 to message our server and the message is handled with handle_cast/2 . In this case we store the new html in the server state and broadcast an event with the new value. On the client side, we subscribe to those events with ctx.handleEvent(event, callback) to update the page accordingly. Event handlers You must eventually register JavaScript handlers for all events that the client may receive. However, the registration can be deferred, if the initialization is asynchronous. For example, the following is perfectly fine: export function init(ctx, data) { fetch(data.someUrl).then((resp) =&gt; { ctx.handleEvent(&quot;update&quot;, (payload) =&gt; { // ... }); }); } Or alternatively: export async function init(ctx, data) { const response = await fetch(data.someUrl); ctx.handleEvent(&quot;update&quot;, (payload) =&gt; { // ... }); } In such case all incoming events are buffered and dispatched once the handler is registered.","ref":"Kino.JS.Live.html","title":"Kino.JS.Live","type":"behaviour"},{"doc":"Makes a synchronous call to the widget server and waits for its reply. See GenServer.call/3 for more details.","ref":"Kino.JS.Live.html#call/3","title":"Kino.JS.Live.call/3","type":"function"},{"doc":"Sends an asynchronous request to the widget server. See GenServer.cast/2 for more details.","ref":"Kino.JS.Live.html#cast/2","title":"Kino.JS.Live.cast/2","type":"function"},{"doc":"Invoked to handle synchronous call/3 messages. See GenServer.handle_call/3 for more details.","ref":"Kino.JS.Live.html#c:handle_call/3","title":"Kino.JS.Live.handle_call/3","type":"callback"},{"doc":"Invoked to handle asynchronous cast/2 messages. See GenServer.handle_cast/2 for more details.","ref":"Kino.JS.Live.html#c:handle_cast/2","title":"Kino.JS.Live.handle_cast/2","type":"callback"},{"doc":"Invoked whenever a new client connects to the server. The returned data is passed to the JavaScript init function of the connecting client.","ref":"Kino.JS.Live.html#c:handle_connect/1","title":"Kino.JS.Live.handle_connect/1","type":"callback"},{"doc":"Invoked to handle client events.","ref":"Kino.JS.Live.html#c:handle_event/3","title":"Kino.JS.Live.handle_event/3","type":"callback"},{"doc":"Invoked to handle all other messages. See GenServer.handle_info/2 for more details.","ref":"Kino.JS.Live.html#c:handle_info/2","title":"Kino.JS.Live.handle_info/2","type":"callback"},{"doc":"Invoked when the widget server started. See GenServer.init/1 for more details.","ref":"Kino.JS.Live.html#c:init/2","title":"Kino.JS.Live.init/2","type":"callback"},{"doc":"Instantiates a live JavaScript widget defined by module . The given init_arg is passed to the init/2 callback when the underlying widget process is started.","ref":"Kino.JS.Live.html#new/2","title":"Kino.JS.Live.new/2","type":"function"},{"doc":"Invoked when the server is about to exit. See GenServer.terminate/2 for more details.","ref":"Kino.JS.Live.html#c:terminate/2","title":"Kino.JS.Live.terminate/2","type":"callback"},{"doc":"","ref":"Kino.JS.Live.html#t:t/0","title":"Kino.JS.Live.t/0","type":"opaque"},{"doc":"State available in Kino.JS.Live server callbacks. Properties :assigns - custom server state kept across callback calls :origin - an opaque identifier of the client that triggered the given action. It is set in Kino.JS.Live.handle_connect/1 and Kino.JS.Live.handle_event/3","ref":"Kino.JS.Live.Context.html","title":"Kino.JS.Live.Context","type":"module"},{"doc":"Stores key-value pairs in the state. Examples assign ( ctx , count : 1 , timestamp : DateTime . utc_now ( ) )","ref":"Kino.JS.Live.Context.html#assign/2","title":"Kino.JS.Live.Context.assign/2","type":"function"},{"doc":"Sends an event to the client. The event is dispatched to the registered JavaScript callback on all connected clients. Examples broadcast_event ( ctx , &quot;new_point&quot; , %{ x : 10 , y : 10 } )","ref":"Kino.JS.Live.Context.html#broadcast_event/3","title":"Kino.JS.Live.Context.broadcast_event/3","type":"function"},{"doc":"Updates an existing key with the given function in the state. Examples assign ( ctx , count : 1 , timestamp : DateTime . utc_now ( ) )","ref":"Kino.JS.Live.Context.html#update/3","title":"Kino.JS.Live.Context.update/3","type":"function"},{"doc":"","ref":"Kino.JS.Live.Context.html#t:origin/0","title":"Kino.JS.Live.Context.origin/0","type":"type"},{"doc":"","ref":"Kino.JS.Live.Context.html#t:t/0","title":"Kino.JS.Live.Context.t/0","type":"type"},{"doc":"A number of output formats supported by Livebook.","ref":"Kino.Output.html","title":"Kino.Output","type":"module"},{"doc":"See control/0 .","ref":"Kino.Output.html#control/1","title":"Kino.Output.control/1","type":"function"},{"doc":"See frame/0 .","ref":"Kino.Output.html#frame/2","title":"Kino.Output.frame/2","type":"function"},{"doc":"See image/0 .","ref":"Kino.Output.html#image/2","title":"Kino.Output.image/2","type":"function"},{"doc":"See input/0 .","ref":"Kino.Output.html#input/1","title":"Kino.Output.input/1","type":"function"},{"doc":"Returns text/0 with the inspected term.","ref":"Kino.Output.html#inspect/2","title":"Kino.Output.inspect/2","type":"function"},{"doc":"See js/0 .","ref":"Kino.Output.html#js/1","title":"Kino.Output.js/1","type":"function"},{"doc":"See markdown/0 .","ref":"Kino.Output.html#markdown/1","title":"Kino.Output.markdown/1","type":"function"},{"doc":"Generates a random binary identifier.","ref":"Kino.Output.html#random_ref/0","title":"Kino.Output.random_ref/0","type":"function"},{"doc":"See text/0 .","ref":"Kino.Output.html#text/1","title":"Kino.Output.text/1","type":"function"},{"doc":"A control widget. All controls have the following properties: :type - one of the recognised control types :ref - a unique identifier :destination - the process to send event messages to On top of that, each control type may have additional attributes. Events All control events are sent to :destination as {:event, id, info} , where info is a map including additional details. In particular, it always includes :origin , which is an opaque identifier of the client that triggered the event.","ref":"Kino.Output.html#t:control/0","title":"Kino.Output.control/0","type":"type"},{"doc":"","ref":"Kino.Output.html#t:control_attrs/0","title":"Kino.Output.control_attrs/0","type":"type"},{"doc":"Outputs placeholder. Frame with type :default includes the initial list of outputs. Other types can be used to update outputs within the given frame. In all cases the outputs order is reversed, that is, most recent outputs are at the top of the stack.","ref":"Kino.Output.html#t:frame/0","title":"Kino.Output.frame/0","type":"type"},{"doc":"","ref":"Kino.Output.html#t:frame_info/0","title":"Kino.Output.frame_info/0","type":"type"},{"doc":"","ref":"Kino.Output.html#t:frame_ref/0","title":"Kino.Output.frame_ref/0","type":"type"},{"doc":"An empty output that should be ignored whenever encountered.","ref":"Kino.Output.html#t:ignored/0","title":"Kino.Output.ignored/0","type":"type"},{"doc":"A raw image in the given format.","ref":"Kino.Output.html#t:image/0","title":"Kino.Output.image/0","type":"type"},{"doc":"An input field. All inputs have the following properties: :type - one of the recognised input types :ref - a unique identifier :id - a persistent input identifier, the same on every reevaluation :label - an arbitrary text used as the input caption :default - the initial input value :destination - the process to send event messages to On top of that, each input type may have additional attributes.","ref":"Kino.Output.html#t:input/0","title":"Kino.Output.input/0","type":"type"},{"doc":"","ref":"Kino.Output.html#t:input_attrs/0","title":"Kino.Output.input_attrs/0","type":"type"},{"doc":"","ref":"Kino.Output.html#t:input_id/0","title":"Kino.Output.input_id/0","type":"type"},{"doc":"JavaScript powered output with dynamic data and events. This output points to a server process that serves data requests and sends custom events. Communication protocol A client process should connect to the server process by sending: { :connect , pid ( ) , info :: %{ ref : ref ( ) , origin : term ( ) } } And expect the following reply: { :connect_reply , initial_data , info :: %{ ref : ref ( ) } } The server process may then keep sending one of the following events: { :event , event :: String . t ( ) , payload :: term ( ) , info :: %{ ref : ref ( ) } } The client process may keep sending one of the following events: { :event , event :: String . t ( ) , payload :: term ( ) , info :: %{ ref : ref ( ) , origin : term ( ) } } See Kino.JS and Kino.JS.Live for more details.","ref":"Kino.Output.html#t:js/0","title":"Kino.Output.js/0","type":"type"},{"doc":"Data describing a custom JS output component. :ref - unique output identifier :pid - the server process holding the data Assets The :assets map includes information about the relevant files. :archive_path - an absolute path to a .tar.gz archive with all the assets :hash - a checksum of all assets in the archive :js_path - a relative asset path pointing to the JavaScript entrypoint module Export The :export map describes how the output should be persisted. The output data is put in a Markdown fenced code block. :info_string - used as the info string for the Markdown code block :key - in case the data is a map and only a specific part should be exported","ref":"Kino.Output.html#t:js_info/0","title":"Kino.Output.js_info/0","type":"type"},{"doc":"Markdown content.","ref":"Kino.Output.html#t:markdown/0","title":"Kino.Output.markdown/0","type":"type"},{"doc":"","ref":"Kino.Output.html#t:ref/0","title":"Kino.Output.ref/0","type":"type"},{"doc":"IO text output, adjacent such outputs are treated as a whole","ref":"Kino.Output.html#t:stdout/0","title":"Kino.Output.stdout/0","type":"type"},{"doc":"Livebook cell output may be one of these values and gets rendered accordingly.","ref":"Kino.Output.html#t:t/0","title":"Kino.Output.t/0","type":"type"},{"doc":"Standalone text block.","ref":"Kino.Output.html#t:text/0","title":"Kino.Output.text/0","type":"type"},{"doc":"Protocol defining term formatting in the context of Livebook.","ref":"Kino.Render.html","title":"Kino.Render","type":"protocol"},{"doc":"Transforms the given value into a Livebook-compatible output.","ref":"Kino.Render.html#to_livebook/1","title":"Kino.Render.to_livebook/1","type":"function"},{"doc":"","ref":"Kino.Render.html#t:t/0","title":"Kino.Render.t/0","type":"type"}]